      program use_PDGEMV
!     ==================
!  +-------------------------------------------------------------------+
!  |  Example program for using the PBLAS routine PDGEMV               |
!  |  for calculating the matrix vector product.                       |
!  |  PDGEMV performs the operation                                    |
!  |  sub(Y) = alpha sub(A) sub(X) + beta sub(Y),                      |
!  |  where sub(A) is a submatrix of a global distributed matrix A     |
!  |  sub(X), sub(Y) are subcolums or subrows of global distribtued    |
!  |  matrices X, Y.                                                   |
!  |  In the example, trans = 'N', incx=incy=1,  and for i=1,m, j=1,n: |
!  |  sub(A)(i,j) = A(i_A+i-1,j_A+j-1)                                 |
!  |  sub(X)(j) = X(i_X+j-1,j_X)                                       |
!  |  sub(Y)(i) = Y(i_Y+i-1,j_Y)                                       |
!  |  The matrices A and X are chosen as                               |
!  |  A(i,j) = mati*i + matj*j, X(i,j) = veci*i + vecj*j               |
!  +-------------------------------------------------------------------+
!
!  all variables should be explicitely declared !
      implicit none 
!  FORTRAN MPI-INCLUDE-file
      include 'mpif.h'

!  parameter constants
      integer           max_al, max_xl, max_yl
      parameter        (max_al=1000000, max_xl=1000000,  max_yl=1000000)
!  variables for definition of matrix A and matrix X
      real*8 mati,matj, veci,vecj
      common/matdat/mati,matj,veci,vecj
!  local variables
      real*8    al(max_al), xl(max_xl), yl(max_yl),
     :          matval, vecval, resval, err
      integer   ctxt, ctxt_sys, ctxt_all,
     :          nproc, nprow, npcol, myid, myrow, mycol,
     :          m_A, n_A, m_X, n_X, m_Y, n_Y, rsrc, csrc,
     :          mb_A, nb_A, mb_X, nb_X, mb_Y, nb_Y,
     :          m_al, n_al, m_xl, n_xl, m_yl, n_yl, 
     :          llda, lldx, lldy, desc_A(9), desc_X(9), desc_Y(9),
     :          m, n, i_A,j_A, i_X, j_X, i_Y, j_Y, incx, incy, 
     :          ipr, ipc, il, jl, i, j, ir, ilin, ncheck, info, 
     :          NUMROC, INDXG2L, INDXG2P, INDXL2G

!  calling BLACS_PINFO(myid,nproc) initializes MPI and returns size and rank
      call BLACS_PINFO(myid,nproc) 
      call BLACS_GET( 0, 0, ctxt_sys )
      ctxt_all = ctxt_sys
      call BLACS_GRIDINIT( ctxt_all, 'C', nproc, 1)
! input parameters
! Parameters for distributed global matrices A, X, Y
      m_A=500; n_A=500; mb_A=24; nb_A=24
      m_X=500; n_X=500; mb_X=24; nb_X=24
      m_Y=500; n_Y=500; mb_Y=24; nb_Y=24
      rsrc=0; csrc=0
! Parameters for submatrix sub(A) and subvectors sub(X), sub(Y)
      m = 300; n = 400
      i_A=10; j_A=1
      i_X=17; j_X=1; incx=1
      i_Y=33; j_Y=11; incy=1
! Parameters for processor grid
      nprow=3; npcol=2
! Set up a process grid of size nprow*npcol
      if (nprow*npcol.gt.nproc) then
         write(6,*) 'nproc = ',nproc,' less then nprow*npcol = '
      end if
      ctxt = ctxt_sys; call BLACS_GRIDINIT( ctxt, 'C', nprow, npcol)
! Processes not belonging to the grid jump to the end of program
      if (ctxt.lt.0) go to 1000
! Get the process coordinates in the grid
      call BLACS_GRIDINFO( ctxt, nprow, npcol, myrow, mycol )
! number of rows and columns of loacal parts al, xl, yl for A, X, Y
      m_al = NUMROC( m_A, mb_A, myrow, rsrc, nprow )
      n_al = NUMROC( n_A, nb_A, mycol, csrc, npcol )
      m_xl = NUMROC( m_X, mb_X, myrow, rsrc, nprow )
      n_xl = NUMROC( n_X, nb_X, mycol, csrc, npcol )
      m_yl = NUMROC( m_Y, mb_Y, myrow, rsrc, nprow )
      n_yl = NUMROC( n_Y, nb_Y, mycol, csrc, npcol )
! Test for sufficient memory
      if (m_al*n_al.gt.max_al.or.
     :    m_xl*n_xl.gt.max_xl.or.
     :    m_yl*n_yl.gt.max_yl )then
       write(6,*)'not enough memory:  al ',m_al*n_al,max_al,
     :                               'xl', m_xl*n_xl,max_xl,
     :                               'yl', m_yl*n_yl,max_yl
       end if 
! set data for A and X; A(i,j) = mati*i + matj*j, X(i,j) = veci*i + vecj*j
      mati = 1.; matj = 1.; veci=2.; vecj = 1.

! initializing descriptors for the distributed matrices A, X, Y:
      llda = max(1,m_al); lldx = max(1,m_xl); lldy = max(1,m_yl)
      call DESCINIT( desc_A, m_A, n_A, mb_A, nb_A, 
     :               rsrc, csrc, ctxt, llda, info )
      call DESCINIT( desc_X, m_X, n_X, mb_X, nb_X,
     :               rsrc, csrc, ctxt, lldx, info )
      call DESCINIT( desc_Y, m_Y, n_Y, mb_Y, nb_Y,
     :               rsrc, csrc, ctxt, lldy, info )

! initialize in parallel the local parts al of A
      do jl = 1 , n_al
         j = INDXL2G(jl,nb_A,mycol,csrc,npcol)
         do il = 1 , m_al
            i = INDXL2G(il,mb_A,myrow,rsrc,nprow)
            al(il+m_al*(jl-1))= matval(i,j)
         end do
      end do
! initialize in parallel the local parts xl of X
      do jl = 1 , n_xl
         j = INDXL2G(jl,nb_X,mycol,csrc,npcol)
         do il = 1 , m_xl
            i = INDXL2G(il,mb_X,myrow,rsrc,nprow)
            xl(il+m_xl*(jl-1))= vecval(i,j)
         end do
      end do

! calculate matrix vector product
      call PDGEMV( 'N',m,n,  1.d0,al,i_A,j_A,desc_A,
     :                            xl,i_X,j_X,desc_X,incx,
     :                       0.d0,yl,i_Y,j_Y,desc_Y,incy )
! check the result 
         err = 0.d0
         ncheck = 10
         do ir = 1 , m
            ilin = i_Y + (j_Y-1)*m_Y + (ir-1)*incy
            j = (ilin-1)/m_Y+1; i = ilin-(j-1)*m_Y
            jl = INDXG2L(j,nb_Y,0,0,npcol)
            ipc = INDXG2P(j,nb_Y,0,csrc,npcol)
            il = INDXG2L(i,mb_Y,0,0,nprow)
            ipr = INDXG2P(i,mb_Y,0,rsrc,nprow)
            if (ipr.eq.myrow.and.ipc.eq.mycol) then
              if(ir.le.ncheck) 
     :         write(6,*) ir, yl(il+lldy*(jl-1)), 
     :             resval(ir,n,i_A,j_A,i_X,j_X,m_X,incx)
               err = err + 
     : abs(yl(il+lldy*(jl-1)) - resval(ir,n,i_A,j_A,i_X,j_X,m_X,incx))
            end if
         end do
         call BLACS_BARRIER(ctxt,'A')
         call sleep(1)
         call dgsum2d( ctxt, 'A', ' ', 1, 1, err, 1, 0, 0 )
         if (myid.eq.0) then
            write(6,*)'error in ax is',err/real(m)
         end if

 1000  continue
      call BLACS_EXIT(0)

      stop
      end

      real*8 function matval(i,j)
      implicit none
      real*8 mati,matj,veci,vecj
      common/matdat/mati,matj,veci,vecj
      integer i,j
      matval = mati*i + matj*j
      return
      end

      real*8 function vecval(i,j)
      implicit none
      real*8 mati,matj,veci,vecj
      common/matdat/mati,matj,veci,vecj
      integer i,j
      vecval = veci*i + vecj*j
      return
      end

      real*8 function resval(i,n,i_A,j_A,i_X,j_X,m_X,incx)
      implicit none
      real*8 mati,matj,veci,vecj
      common/matdat/mati,matj,veci,vecj
      integer i, n, i_A, j_A, i_X, j_X, m_X, incx,
     :        j, jlin, ix, jx
      real*8 a1, a2, x1, x2, matval, vecval, res
      
      res = 0.0d0
      do j = 1 , n
         jlin = i_X + m_X*(j_X-1) + (j-1)*incx
         jx = (jlin-1)/m_X+1; ix = jlin - (jx-1)*m_X
         res = res + matval(i_A+i-1,j_A+j-1)*vecval(ix,jx)
      end do
      resval = res

      return
      end


