      program use_PDGESV
!     ==================
!  +-------------------------------------------------------------------+
!  |  Example program for using the ScaLAPACK routine PDGESV for       |
!  |  solving a system of linear equation with nrhs right hand sides.  | 
!  |  PDGESV performs the operation                                    |
!  |  a x = b                                                          |
!  |  where the coefficient matrix is given as a submatrix sub(A)      |
!  |  of size nxn of a global distributed matrix A,                    |
!  |  the right hand sides as a submatrix sub(B) of size nxnrhs of a   |
!  |  global distribtued matrix B and the nrhs solution vectors        |
!  |  overwrite values of the right hand sides in sub(B).              |
!  |                                                                   |
!  |  sub(A)(1:n,1:n) = A(i_A:i_A+n-1,j_A:j_A+n-1)                     |
!  |  sub(B)(1:n,1:nrhs) = B(i_B:i_B+n-1,j_B:j_B+nrhs-1)               |
!  |  The the global matrices A and B are chosen as                    |
!  |  A(i,j) = 1./(mati + matj*abs(i-j)), B(i,j) = bi*i + bj*j         |
!  +-------------------------------------------------------------------+
!
      implicit none 

!  parameter constants
      integer           max_al, max_bl, max_vl
      parameter        (max_al=10000, max_bl=10000,  max_vl=10000)
!  variables for definition of matrix A and matrix X
      real*8 mati,matj,bi,bj
      common/matdat/mati,matj,bi,bj
!  local variables
      real*8    al(max_al), al_s(max_al), bl(max_bl), rl(max_bl),
     :          matval, bval, err
      integer   ctxt, ctxt_sys, ctxt_all,
     :          nproc, nprow, npcol, myid, myrow, mycol,
     :          m_A, n_A, m_B, n_B, rsrc_A, csrc_A, rsrc_B, csrc_B,
     :          mb_A, nb_A, mb_B, nb_B,
     :          m_al, n_al, m_bl, n_bl, m_cl, 
     :          llda, lldb, lldc, desc_A(9), desc_B(9), desc_R(9),
     :          n, nrhs, ncheck, i_A, j_A, i_B, j_B, 
     :          irhs, ipr, ipc, il, jl, i, j, info, vl(max_vl),  
     :          NUMROC, INDXL2G, INDXG2L, INDXG2P
!  calling BLACS_PINFO(myid,nproc) initializes MPI and returns size and rank
      call BLACS_PINFO(myid,nproc) 
      call BLACS_GET( 0, 0, ctxt_sys )
      ctxt_all = ctxt_sys
      call BLACS_GRIDINIT( ctxt_all, 'C', nproc, 1)
! input parameters
! Parameters for distributed global matrices A, B
! Restrictions are: 
!   mb_A = nb_A, mb_B = nb_A
      m_A=10; n_A=10; mb_A=2; nb_A=2
      m_B=13; n_B=4; mb_B=2; nb_B=3
      rsrc_A=0; csrc_A=1; rsrc_B=1; csrc_B=0
! Parameters for submatrices sub(A) and sub(B)
! Restrictions are:
!   i_A+n-1 <= m_A, j_A+n-1<=n_A, i_B+n-1<=m_B, j_B+nrhs-1<=n_B
!   mod(j_A-1,nb_A) = 0,  mod(i_A-1,mb_A) = mod(i_B-1,mb_B)
!   the process row containing the row i_A of A 
!   must also contain the row i_B of B
      n = 9; nrhs = 3
      i_A=1; j_A=1
      i_B=3; j_B=1
! Parameters for processor grid
      nprow=2; npcol=3
! Set up a process grid of size nprow*npcol
      if (nprow*npcol.gt.nproc) then
         write(6,*) 'nproc = ',nproc,' less then nprow*npcol = '
      end if
      ctxt = ctxt_sys; call BLACS_GRIDINIT( ctxt, 'C', nprow, npcol)
! Processes not belonging to the grid jump to the end of program
      if (ctxt.lt.0) go to 1000
! Get the process coordinates in the grid
      call BLACS_GRIDINFO( ctxt, nprow, npcol, myrow, mycol )
! number of rows and columns of loacal parts al, bl for A, B
      m_al = NUMROC( m_A, mb_A, myrow, rsrc_A, nprow )
      n_al = NUMROC( n_A, nb_A, mycol, csrc_A, npcol )
      m_bl = NUMROC( m_B, mb_B, myrow, rsrc_B, nprow )
      n_bl = NUMROC( n_B, nb_B, mycol, csrc_B, npcol )
! Test for sufficient memory
      if (m_al*n_al.gt.max_al.or.
     :    m_bl*n_bl.gt.max_bl) then
       write(6,*)'not enough memory:  al ',m_al*n_al,max_al,
     :                               'bl', m_bl*n_bl,max_bl
       end if 
! set data for A and B; A(i,j) = 1./(mati+matj*(abs(i-j)), B(i,j) = bi*i+bj*j
      mati = 1.; matj = 5.; bi=1.; bj = 1.

! initializing descriptors for the distributed matrices A, B:
      llda = max(1,m_al); lldb = max(1,m_bl); lldc = max(1,m_cl)
      call DESCINIT( desc_A, m_A, n_A, mb_A, nb_A, 
     :               rsrc_A, csrc_A, ctxt, llda, info )
      call DESCINIT( desc_B, m_B, n_B, mb_B, nb_B,
     :               rsrc_B, csrc_B, ctxt, lldb, info )

! initialize in parallel the local parts of A in al and in al_s
      do jl = 1 , n_al
         j = INDXL2G(jl,nb_A,mycol,csrc_A,npcol)
         do il = 1 , m_al
            i = INDXL2G(il,mb_A,myrow,rsrc_A,nprow)
            al(il+m_al*(jl-1))= matval(i,j)
            al_s(il+m_al*(jl-1))= matval(i,j)
         end do
      end do
! initialize in parallel the local parts of B in bl
      do jl = 1 , n_bl
         j = INDXL2G(jl,nb_B,mycol,csrc_B,npcol)
         do il = 1 , m_bl
            i = INDXL2G(il,mb_B,myrow,rsrc_B,nprow)
            bl(il+m_bl*(jl-1))= bval(i,j)
         end do
      end do
!      call inspect(bl,'bl',desc_B) 

! solve the systems of equations
      call PDGESV( n, nrhs, al, i_A, j_A, desc_A, vl,
     :                      bl, i_B, j_B, desc_B, info)
!      call inspect(bl,'x',desc_B)
! check the result
      ncheck = 10
      do irhs = 1 , nrhs
         call BLACS_BARRIER(ctxt,'A')
         call sleep(1)
         if (myid.eq.0) write(6,*)'solution vector x for irhs =',irhs
         call BLACS_BARRIER(ctxt,'A')
         call sleep(1)
         jl = INDXG2L(irhs,nb_B,mycol,csrc_B,npcol)
         ipc = INDXG2P(irhs,nb_B,mycol,csrc_B,npcol)
         do i = 1 , min(ncheck,n)
            il = INDXG2L(i_B+i-1,mb_B,myrow,rsrc_B,nprow)
            ipr = INDXG2P(i_B+i-1,mb_B,myrow,rsrc_B,nprow)
            if (ipr.eq.myrow.and.ipc.eq.mycol) write(6,*)
     :             i,bl(il+m_bl*(jl-1))
         end do
         call BLACS_BARRIER(ctxt,'A')
         call sleep(1)
         if (myid.eq.0) write(6,*)'check of rhs vector for irhs =',irhs
         call BLACS_BARRIER(ctxt,'A')
         call sleep(1)
! matrix vector multiplication of coefficient matrix in al_s 
! times solutionvector in bl
! set up descriptor for resultvector R 
         call DESCINIT( desc_R, n, 1, n, 1,
     :               0, 0, ctxt, n, info )
         call PDGEMV( 'N',n,n,  1.d0,al_s,i_A,j_A,desc_A,
     :                               bl,i_B,irhs,desc_B,1,
     :                          0.d0,rl,1,1,desc_R,1, ) 
         err = 0.d0
         do i = 1 , n
            if (0.eq.myrow.and.0.eq.mycol.and.i.le.ncheck) write(6,*)
     :             i,rl(i), bval(i_B+i-1,j_B+irhs-1)
            err = err + abs(rl(i)- bval(i_B+i-1,j_B+irhs-1))
         end do
         call BLACS_BARRIER(ctxt,'A')
         call sleep(1)
         if (myid.eq.0) then
            write(6,*)'error in b-ax is',err/real(n)
         end if         

       end do

 1000  continue
      call BLACS_EXIT(0)

      stop
      end

      subroutine inspect(a,na,desc)
! Inspect the local elements of the distributed matrix on each process
      implicit none

      character*5 na
      real*8 a(*)
      integer desc(9), ctxt, nprow ,npcol, myrow, mycol,
     :        m, n, mb, nb, m_a, n_a, rsrc, csrc, ipc, ipr, i, j,
     :        NUMROC

      ctxt = desc(2)

      call BLACS_GRIDINFO( ctxt, nprow, npcol, myrow, mycol )
      if (myrow.eq.0.and.mycol.eq.0) then
         write(6,*)'inspect  ',na
      end if
      call BLACS_BARRIER(ctxt,'A')
      call sleep(1)
      m = desc(3); n = desc(4); mb = desc(5); nb = desc(6)
      rsrc = desc(7); csrc = desc(8)
      m_a = NUMROC( m, mb, myrow, rsrc, nprow )
      n_a = NUMROC( n, nb, mycol, csrc, npcol )

      do ipc = 0 , npcol-1
         do ipr = 0, nprow-1
            if (ipr.eq.myrow.and.ipc.eq.mycol) then
                write(6,232) ipr, ipc
                do i = 1 , m_a
                   write(6,233)(a(i+m_a*(j-1)),j = 1,n_a)
                end do
 232   format(' on process (',i2,','i2,')')
 233   format(10f8.4)
            end if
            call BLACS_BARRIER(ctxt,'A')
            call sleep(1)
         end do
      end do

      return
      end

      real*8 function matval(i,j)
      implicit none
      real*8 mati,matj,bi,bj
      common/matdat/mati,matj,bi,bj
      integer i,j
!      matval = i+100*j
      matval = 1./(mati + matj*abs(i-j))
      return
      end

      real*8 function bval(i,j)
      implicit none
      real*8 mati,matj,bi,bj
      common/matdat/mati,matj,bi,bj
      integer i,j
      bval = bi*i + bj*j
      return
      end
